use std::time::Duration;

#[cfg(feature = "f32")]
pub type Float = f32;
#[cfg(feature = "f32")]
pub use std::f32 as floats;
#[cfg(not(feature = "f32"))]
pub type Float = f64;
#[cfg(not(feature = "f32"))]
pub use std::f64 as floats;

#[derive(Debug, Copy, Clone, Eq, PartialEq)]
pub(crate) enum Mode {
    Once,
    RepeatFrom(usize),
    PingPong,
}

impl Default for Mode {
    fn default() -> Self {
        Self::RepeatFrom(0usize)
    }
}

pub struct Frame {
    pub index: usize,
    pub duration: Duration,
}

impl Frame {
    pub fn new(index: usize, duration: Duration) -> Self {
        assert!(!duration.is_zero(), "must be a non-zero duration");
        Self { index, duration }
    }
}

pub struct Animation {
    pub(crate) frames: Vec<Frame>,
    pub(crate) mode: Mode,
}

impl Animation {
    pub fn len(&self) -> usize {
        self.frames.len()
    }

    pub fn is_empty(&self) -> bool {
        self.len() == 0
    }

    pub fn from_indices(indices: impl Iterator<Item = usize>, frame_rate: FrameRate) -> Self {
        let duration = if frame_rate.is_total_duration {
            frame_rate.frame_duration.div_f64(indices.cloned().count() as f64)
        } else {
            frame_rate.frame_duration
        };

        indices
            .map(|index| Frame::new(index, duration))
            .collect()
    }
}

#[derive(Debug, Copy, Clone, Eq, PartialEq)]
pub struct FrameRate {
    frame_duration: Duration,
    is_total_duration: bool,
}

impl FrameRate {
    pub fn from_fps(fps: Float) -> Self {
        assert!(fps.is_finite(), "FPS must be finite");
        Self::from_total_duration(Duration::from_secs(1).div_f64(fps))
    }

    pub fn from_frame_duration(frame_duration: Duration) -> Self {
        Self {
            frame_duration,
            is_total_duration: false,
        }
    }

    pub fn from_total_duration(frame_duration: Duration) -> Self {
        Self {
            frame_duration,
            is_total_duration: true,
        }
    }
}

#[derive(Debug, Default, Clone)]
pub struct State {
    animation_index: usize,
    sprite_index: usize,
    elapsed_time: Duration,
    is_reverse: bool,
    is_complete: bool,
}

impl State {
    pub fn new() -> Self {
        Self::default()
    }

    pub fn frame<'a>(&self, animation: &'a Animation) -> &'a Frame {
        &animation.frames[self.animation_index % animation.len()]
    }

    pub fn update(&mut self, animation: &Animation, delta: Duration) {
        let mut current_frame = self.frame(animation);
        self.sprite_index = current_frame.index;
        self.elapsed_time += delta;
        while self.elapsed_time >= current_frame.duration {
            let is_last_frame = self.animation_index >= animation.len();

            match animation.mode {
                Mode::Once if self.is_complete => {}
                Mode::Once if is_last_frame => self.is_complete = true,
                Mode::Once => self.animation_index += 1,
                Mode::RepeatFrom(start_frame) if is_last_frame => {
                    self.animation_index = start_frame
                }
                Mode::RepeatFrom(_start_frame) => self.animation_index += 1,
                Mode::PingPong if self.is_reverse && self.animation_index == 0 => {
                    self.is_reverse = false;
                    self.animation_index += 1;
                }
                Mode::PingPong if is_last_frame => {
                    self.is_reverse = true;
                    self.animation_index += 1;
                }
                Mode::PingPong => self.animation_index += 1,
            }

            self.elapsed_time -= current_frame.duration;
            current_frame = self.frame(animation);
            self.sprite_index = current_frame.index;
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use rstest::{fixture, rstest};

    #[rstest]
    #[should_panic(expected = "must be a non-zero duration")]
    fn should_panic_on_zero_duration_frame(#[values(0.0)] value: Float) {
        let _ = Frame::new(0usize, Duration::from_secs_f64(value));
    }

    #[fixture]
    fn state() -> State {
        State::new()
    }

    #[fixture]
    fn sixy_fps_4_frame_animation() -> Animation {
        Animation::f
    }

    #[rstest]
    fn update_sixty_fps(sixy_fps: Duration) {}
}
