use bevy::prelude::*;
use bevy_ecs_ldtk::prelude::*;
use bevy_smooth_pixel_camera::{components::PixelCamera, viewport::ViewportSize};

#[derive(Component, Debug)]
pub struct Follow {
    pub target: Entity,
    pub offset: Vec3,
}

impl Follow {
    pub fn new(target: Entity) -> Self {
        Self {
            target,
            offset: Vec3::ZERO,
        }
    }

    pub fn with_offset(self, offset: Vec3) -> Self {
        Self { offset, ..self }
    }
}

// link this to the camera's target
// https://trouv.github.io/bevy_ecs_ldtk/latest/how-to-guides/create-bevy-relations-from-ldtk-entity-references.html
#[derive(Debug, Default, Deref, DerefMut, Component)]
pub struct UnresolvedTargetRef(Option<EntityIid>);
impl UnresolvedTargetRef {
    pub fn from_field(instance: &EntityInstance) -> UnresolvedTargetRef {
        Self(
            instance
                .get_maybe_entity_ref_field("Target")
                .expect("expected entity to have mother entity ref field")
                .as_ref()
                .map(|entity_ref| EntityIid::new(entity_ref.entity_iid.clone())),
        )
    }
}

#[derive(Bundle)]
pub struct PixelCameraBundle {
    pixel_camera: PixelCamera,
    camera_2d_bundle: Camera2dBundle,
}

impl Default for PixelCameraBundle {
    fn default() -> Self {
        Self {
            pixel_camera: PixelCamera::from_size(ViewportSize::PixelFixed(1)),
            camera_2d_bundle: Camera2dBundle::default(),
        }
    }
}

#[derive(Bundle, Default, LdtkEntity)]
pub struct CameraEntity {
    #[with(UnresolvedTargetRef::from_field)]
    unresolved_target_ref: UnresolvedTargetRef,
    pixel_camera_bundle: PixelCameraBundle,
}

pub struct CameraPlugin;
impl Plugin for CameraPlugin {
    fn build(&self, app: &mut App) {
        app.add_systems(
            Update,
            (follow_target, resolve_target_references, dynamic_scale),
        )
        .register_ldtk_entity::<CameraEntity>("Camera");
    }
}

fn follow_target(
    mut follower_q: Query<(&mut PixelCamera, &Follow)>,
    target_q: Query<&GlobalTransform, Without<Follow>>,
) {
    for (mut camera, follow) in follower_q.iter_mut() {
        if let Ok(target) = target_q.get(follow.target) {
            let result = target.transform_point(follow.offset);
            camera.subpixel_pos = Vec2::new(result.x, result.y);
        }
    }
}

pub fn resolve_target_references(
    mut commands: Commands,
    unresolved_targets: Query<(Entity, &UnresolvedTargetRef), Added<UnresolvedTargetRef>>,
    ldtk_entities: Query<(Entity, &EntityIid)>,
) {
    for (child_entity, unresolved_target_ref) in unresolved_targets.iter() {
        if let Some(target_iid) = unresolved_target_ref.0.as_ref() {
            let (target_entity, _) = ldtk_entities
                .iter()
                .find(|(_, iid)| *iid == target_iid)
                .expect("enemy's target entity should exist");

            commands
                .entity(child_entity)
                .remove::<UnresolvedTargetRef>()
                .insert(Follow::new(target_entity));
        } else {
            commands
                .entity(child_entity)
                .remove::<UnresolvedTargetRef>();
        }
    }
}

pub fn dynamic_scale(mut query: Query<&mut OrthographicProjection>) {
    for mut projection in query.iter_mut() {
        projection.scale = 0.5;
    }
}
