use bevy::prelude::*;
use bevy_ecs_ldtk::prelude::*;
use bevy_xpbd_2d::prelude::*;

#[derive(Component, Default, Reflect)]
pub struct ZIndex(i32);

impl ZIndex {
    pub fn from_field(instance: &EntityInstance) -> ZIndex {
        Self(
            *instance
                .get_int_field("ZIndex")
                .expect("expected entity to have Z Index field"),
        )
    }
}

impl<T: Into<i32>> From<T> for ZIndex {
    fn from(value: T) -> Self {
        ZIndex(value.into())
    }
}

#[derive(Component, Default)]
pub struct Tile;

#[derive(Bundle)]
pub struct ColliderBundle {
    collider: Collider,
    rigidbody: RigidBody,
}

impl Default for ColliderBundle {
    fn default() -> Self {
        Self {
            collider: Collider::rectangle(16., 16.),
            rigidbody: RigidBody::Static,
        }
    }
}

impl From<&EntityInstance> for ColliderBundle {
    fn from(value: &EntityInstance) -> Self {
        let rigidbody = match value.get_enum_field("Collider") {
            Ok(_) => RigidBody::Kinematic,
            _ => RigidBody::Static,
        };

        Self {
            rigidbody,
            collider: Collider::rectangle(value.width as f32, value.height as f32),
        }
    }
}

#[derive(Default, Bundle, LdtkIntCell)]
pub struct TileIntCell {
    tile: Tile,
    collider_bundle: ColliderBundle,
}

#[derive(Default, Bundle, LdtkEntity)]
pub struct PlatformEntity {
    #[with(ZIndex::from_field)]
    z_index: ZIndex,
    #[from_entity_instance]
    collider_bundle: ColliderBundle,
    #[sprite_sheet_bundle]
    sprite_sheet_bundle: SpriteSheetBundle,
}

pub struct LevelPlugin;
impl Plugin for LevelPlugin {
    fn build(&self, app: &mut App) {
        app.register_type::<ZIndex>()
            .add_systems(Last, update_z_index)
            .register_ldtk_entity::<PlatformEntity>("OneWayPlatform")
            .register_default_ldtk_int_cell_for_layer::<TileIntCell>("Ground");
    }
}

fn update_z_index(mut query: Query<(&mut Transform, &ZIndex), Changed<ZIndex>>) {
    for (mut transform, z_index) in query.iter_mut() {
        transform.translation.z = z_index.0 as f32;
    }
}
