mod controller;
mod input;
mod movement;

use bevy::prelude::*;
use bevy_ecs_ldtk::prelude::*;
use leafwing_input_manager::{action_state::ActionState, InputManagerBundle};

use self::{
    controller::{CharacterControllerBundle, CharacterControllerPlugin, Jumping},
    input::{InputPlugin, PlayerAction},
    movement::MovementDirection,
};
use crate::{animation::*, level::ZIndex};


#[derive(Component, Default)]
pub struct Player;

#[derive(Bundle)]
pub struct PlayerBundle {
    player: Player,
    input_manager_bundle: InputManagerBundle<PlayerAction>,
    character_controller_bundle: CharacterControllerBundle,
    animation_bundle: AnimationBundle,
}

impl Default for PlayerBundle {
    fn default() -> Self {
        Self {
            player: Player,
            input_manager_bundle: InputManagerBundle::with_map(PlayerAction::default_input_map()),
            character_controller_bundle: CharacterControllerBundle::default(),
            animation_bundle: AnimationBundle::from_animation(benimator::Animation::from_indices(
                0..4,
                benimator::FrameRate::from_fps(12.),
            )),
        }
    }
}

#[derive(Default, Bundle, LdtkEntity)]
pub struct PlayerEntity {
    player_bundle: PlayerBundle,
    #[with(ZIndex::from_field)]
    z_index: ZIndex,
    #[sprite_sheet_bundle]
    sprite_sheet_bundle: SpriteSheetBundle,
}

pub struct PlayerPlugin;
impl Plugin for PlayerPlugin {
    fn build(&self, app: &mut App) {
        app.add_plugins((InputPlugin, CharacterControllerPlugin))
            .add_systems(Update, (apply_movement_input, apply_jumping_input))
            .register_ldtk_entity::<PlayerEntity>("Player");
    }
}

fn apply_movement_input(mut query: Query<(&ActionState<PlayerAction>, &mut MovementDirection)>) {
    for (action_state, mut direction) in query.iter_mut() {
        direction.0 = action_state
            .clamped_axis_pair(&PlayerAction::Move)
            .map_or(0., |v| v.x())
    }
}

fn apply_jumping_input(
    mut commands: Commands,
    mut query: Query<(Entity, &ActionState<PlayerAction>, Has<Jumping>)>,
) {
    for (entity, action_state, is_jumping) in query.iter_mut() {
        let wants_to_jump = action_state.pressed(&PlayerAction::Jump);
        if wants_to_jump && !is_jumping {
            commands.entity(entity).insert(Jumping);
        } else if !wants_to_jump && is_jumping {
            commands.entity(entity).remove::<Jumping>();
        }
    }
}
