use std::ops::Mul;

use bevy::prelude::*;
use bevy_xpbd_2d::math::{Scalar, PI};

pub trait DefaultMultiple {
    fn default_multiple(n: Scalar) -> Self;
}

impl<T: Default + Mul<Scalar, Output = T>> DefaultMultiple for T {
    fn default_multiple(n: Scalar) -> Self {
        Self::default() * n
    }
}

#[derive(Component, Reflect, Deref)]
pub struct JumpImpulse(pub Scalar);

impl Default for JumpImpulse {
    fn default() -> Self {
        JumpImpulse(7.)
    }
}

impl Mul<Scalar> for JumpImpulse {
    type Output = Self;

    fn mul(self, rhs: Scalar) -> Self::Output {
        Self(self.0 * rhs)
    }
}

#[derive(Component, Reflect, Deref)]
pub struct MaxJumpDuration(pub Scalar);

impl Default for MaxJumpDuration {
    fn default() -> Self {
        MaxJumpDuration(0.1)
    }
}

impl Mul<Scalar> for MaxJumpDuration {
    type Output = Self;

    fn mul(self, rhs: Scalar) -> Self::Output {
        Self(self.0 * rhs)
    }
}

#[derive(Component, Deref, Reflect)]
pub struct MovementAcceleration(pub Scalar);

impl Mul<Scalar> for MovementAcceleration {
    type Output = Self;

    fn mul(self, rhs: Scalar) -> Self::Output {
        Self(self.0 * rhs)
    }
}

impl Default for MovementAcceleration {
    fn default() -> Self {
        MovementAcceleration(75.)
    }
}

#[derive(Component, Deref, Reflect)]
pub struct MovementDampingFactor(pub Scalar);

impl Mul<Scalar> for MovementDampingFactor {
    type Output = Self;

    fn mul(self, rhs: Scalar) -> Self::Output {
        Self(self.0 * rhs)
    }
}

impl Default for MovementDampingFactor {
    fn default() -> Self {
        MovementDampingFactor(0.9)
    }
}

#[derive(Component, Deref, Reflect)]
pub struct MaxSlopeAngle(pub Scalar);

impl Mul<Scalar> for MaxSlopeAngle {
    type Output = Self;

    fn mul(self, rhs: Scalar) -> Self::Output {
        Self(self.0 * rhs)
    }
}

impl Default for MaxSlopeAngle {
    fn default() -> Self {
        MaxSlopeAngle(PI * 0.45)
    }
}

#[derive(Component, Default, Reflect)]
pub struct MovementDirection(pub Scalar);

#[derive(Bundle, Default)]
pub struct MovementBundle {
    pub acceleration: MovementAcceleration,
    pub damping: MovementDampingFactor,
    pub jump_impulse: JumpImpulse,
    pub max_slope_angle: MaxSlopeAngle,
    pub direction: MovementDirection,
}

impl DefaultMultiple for MovementBundle {
    fn default_multiple(n: Scalar) -> Self {
        Self {
            acceleration: MovementAcceleration::default_multiple(n),
            damping: MovementDampingFactor::default_multiple(n),
            jump_impulse: JumpImpulse::default_multiple(n),
            max_slope_angle: MaxSlopeAngle::default_multiple(n),
            direction: MovementDirection::default(),
        }
    }
}

